@@ -1,5 +1,3 @@
-from __future__ import absolute_import
-
 import datetime
 import inspect
 
@@ -10,33 +8,40 @@ from . import exceptions, utils
 
 class Granularity(object):
     class _Base(object):
+        # To be overridden.
+        _order = None
+        _key = None
+        _round_rule = None
+        _name = None
+        _duration = None
+
         class _BaseMetaclass(type):
-            def __lt__(self, other):
-                return self._order < other._order
+            def __lt__(cls, other):
+                return cls._order < other._order
 
-            def __gt__(self, other):
-                return self._order > other._order
+            def __gt__(cls, other):
+                return cls._order > other._order
 
-            def __le__(self, other):
-                return self._order <= other._order
+            def __le__(cls, other):
+                return cls._order <= other._order
 
-            def __ge__(self, other):
-                return self._order >= other._order
+            def __ge__(cls, other):
+                return cls._order >= other._order
 
-            def __eq__(self, other):
-                return self._order == other._order
+            def __eq__(cls, other):
+                return cls._order == other._order
 
-            def __str__(self):
-                return self._name
+            def __str__(cls):
+                return cls._name
 
         __metaclass__ = _BaseMetaclass
 
         @utils.class_property
-        def name(cls):
+        def name(cls): # pylint: disable=no-self-argument
             return cls._name
 
         @utils.class_property
-        def key(cls):
+        def key(cls): # pylint: disable=no-self-argument
             return cls._key
 
         @classmethod
@@ -111,7 +116,7 @@ class Granularity(object):
         _duration = 86400
 
     @utils.class_property
-    def values(cls):
+    def values(cls): # pylint: disable=no-self-argument
         if not hasattr(cls, '_values'):
             cls._values = tuple(sorted(
                 [
@@ -137,64 +142,37 @@ assert Granularity.Seconds > Granularity.Seconds10 > Granularity.Minutes > Granu
 
 class Stream(object):
     def __init__(self, all_tags):
-        tags = []
-        for tag in all_tags:
-            try:
-                self.id = tag['stream_id']
-                continue
-            except (ValueError, KeyError, TypeError):
-                pass
+        self.tags = all_tags.copy()
 
-            try:
-                self.value_downsamplers = tag['value_downsamplers']
-                continue
-            except (ValueError, KeyError, TypeError):
-                pass
+        try:
+            self.id = all_tags['stream_id']
+            self.value_downsamplers = all_tags['value_downsamplers']
+            self.time_downsamplers = all_tags['time_downsamplers']
+            self.highest_granularity = all_tags['highest_granularity']
+            self.pending_backprocess = bool(all_tags['pending_backprocess'])
+            self.latest_datapoint = all_tags['latest_datapoint']
+            self.earliest_datapoint = all_tags['earliest_datapoint']
+            self.downsampled_until = all_tags['downsampled_until']
+            self.value_type = all_tags['value_type']
 
-            try:
-                self.time_downsamplers = tag['time_downsamplers']
-                continue
-            except (ValueError, KeyError, TypeError):
-                pass
+            if 'derived_from' in all_tags:
+                self.derived_from = all_tags['derived_from']
+                del self.tags['derived_from']
+            if 'contributes_to' in all_tags:
+                self.contributes_to = all_tags['contributes_to']
+                del self.tags['contributes_to']
 
-            try:
-                self.highest_granularity = tag['highest_granularity']
-                continue
-            except (ValueError, KeyError, TypeError):
-                pass
-
-            try:
-                self.derived_from = tag['derived_from']
-                continue
-            except (ValueError, KeyError, TypeError):
-                pass
-
-            try:
-                self.contributes_to = tag['contributes_to']
-                continue
-            except (ValueError, KeyError, TypeError):
-                pass
-
-            try:
-                self.pending_backprocess = bool(tag['pending_backprocess'])
-                continue
-            except (ValueError, KeyError, TypeError):
-                pass
-
-            tags.append(tag)
-
-        self.tags = tags
-
-        if not hasattr(self, 'id'):
-            raise ValueError("Supplied tags are missing 'stream_id'.")
-        if not hasattr(self, 'value_downsamplers'):
-            raise ValueError("Supplied tags are missing 'value_downsamplers'.")
-        if not hasattr(self, 'time_downsamplers'):
-            raise ValueError("Supplied tags are missing 'time_downsamplers'.")
-        if not hasattr(self, 'highest_granularity'):
-            raise ValueError("Supplied tags are missing 'highest_granularity'.")
-        if not hasattr(self, 'pending_backprocess'):
-            raise ValueError("Supplied tags are missing 'pending_backprocess'.")
+            del self.tags['stream_id']
+            del self.tags['value_downsamplers']
+            del self.tags['time_downsamplers']
+            del self.tags['highest_granularity']
+            del self.tags['pending_backprocess']
+            del self.tags['latest_datapoint']
+            del self.tags['earliest_datapoint']
+            del self.tags['downsampled_until']
+            del self.tags['value_type']
+        except KeyError, exception:
+            raise ValueError("Supplied tags are missing %s." % exception)
 
 RESERVED_TAGS = (
     'stream_id',
@@ -204,11 +182,14 @@ RESERVED_TAGS = (
     'derived_from',
     'contributes_to',
     'pending_backprocess',
+    'latest_datapoint',
+    'earliest_datapoint',
+    'downsampled_until',
+    'value_type',
 )
 
 VALUE_DOWNSAMPLERS = {
     'mean': 'm', # average of all datapoints
-    'median': 'e', # median of all datapoints
     'sum': 's', # sum of all datapoints
     'min': 'l', # minimum value of all dataponts (key mnemonic: l for lower)
     'max': 'u', # maximum value of all datapoints (key mnemonic: u for upper)
@@ -223,11 +204,9 @@ VALUE_DOWNSAMPLERS = {
 # Count of timestamps is the same as count of values
 TIME_DOWNSAMPLERS = {
     'mean': 'm', # average of all timestamps
-    'median': 'e', # median of all timestamps
     'first': 'a', # the first timestamp of all datapoints (key mnemonic: a is the first in the alphabet)
     'last': 'z', # the last timestamp of all datapoints (key mnemonic: z is the last in the alphabet)
     'intervals_mean': 'i', # average of all interval lengths (key mnemonic: i for interval)
-    'intervals_median': 'n', # median of all interval lengths (key mnemonic: mediaN)
     'intervals_min': 'l', # minimum of all interval lengths (key mnemonic: l for lower)
     'intervals_max': 'u', # maximum of all interval lengths (key mnemonic: u for upper)
     'intervals_sum_squares': 'q', # sum of squares of all interval lengths
@@ -241,9 +220,14 @@ DERIVE_OPERATORS = {
     'counter_derivative': 'COUNTER_DERIVATIVE', # derivative of a monotonically increasing counter stream
 }
 
+VALUE_TYPES = {
+    'numeric',
+    'graph',
+}
 
-class Datapoints(object):
-    def batch_size(self, batch_size):
+
+class ResultsBase(object):
+    def batch_size(self, batch_size): # pylint: disable=unused-argument
         # Ignore by default, this is just for optimization
         return
 
@@ -260,6 +244,14 @@ class Datapoints(object):
         raise NotImplementedError
 
 
+class Streams(ResultsBase):
+    pass
+
+
+class Datapoints(ResultsBase):
+    pass
+
+
 class Datastream(object):
     Granularity = Granularity
     Stream = Stream
@@ -270,19 +262,19 @@ class Datastream(object):
 
     def __init__(self, backend):
         """
-        Class constructor.
+        Initializes the Datastream API.
 
         :param backend: Backend instance
         """
 
         self.backend = backend
 
-    def ensure_stream(self, query_tags, tags, value_downsamplers, highest_granularity, derive_from=None, derive_op=None, derive_args=None):
+    def ensure_stream(self, query_tags, tags, value_downsamplers, highest_granularity, derive_from=None, derive_op=None, derive_args=None, value_type=None):
         """
         Ensures that a specified stream exists.
 
-        :param query_tags: Tags which uniquely identify a stream
-        :param tags: Tags that should be used (together with `query_tags`) to create a
+        :param query_tags: A dictionary of tags which uniquely identify a stream
+        :param tags: A dictionary of tags that should be used (together with `query_tags`) to create a
                      stream when it doesn't yet exist
         :param value_downsamplers: A set of names of value downsampler functions for this stream
         :param highest_granularity: Predicted highest granularity of the data the stream
@@ -290,6 +282,7 @@ class Datastream(object):
         :param derive_from: Create a derivate stream
         :param derive_op: Derivation operation
         :param derive_args: Derivation operation arguments
+        :param value_type: Optional value type (defaults to `numeric`)
         :return: A stream identifier
         """
 
@@ -310,14 +303,19 @@ class Datastream(object):
             if derive_args is None:
                 derive_args = {}
 
-        return self.backend.ensure_stream(query_tags, tags, value_downsamplers, highest_granularity, derive_from, derive_op, derive_args)
+        if value_type is None:
+            value_type = 'numeric'
+        if value_type not in VALUE_TYPES:
+            raise exceptions.UnsupportedValueType("Unsupported value type: %s" % value_type)
+
+        return self.backend.ensure_stream(query_tags, tags, value_downsamplers, highest_granularity, derive_from, derive_op, derive_args, value_type)
 
     def get_tags(self, stream_id):
         """
         Returns the tags for the specified stream.
 
         :param stream_id: Stream identifier
-        :return: A list of tags for the stream
+        :return: A dictionary of tags for the stream
         """
 
         return self.backend.get_tags(stream_id)
@@ -327,17 +325,21 @@ class Datastream(object):
         Updates stream tags with new tags, overriding existing ones.
 
         :param stream_id: Stream identifier
-        :param tags: A list of new tags
+        :param tags: A dictionary of new tags
         """
 
+        for key in tags.keys():
+            if key in RESERVED_TAGS:
+                raise exceptions.ReservedTagNameError
+
         self.backend.update_tags(stream_id, tags)
 
     def remove_tag(self, stream_id, tag):
         """
-        Removes stream tag.
+        Removes a stream tag.
 
         :param stream_id: Stream identifier
-        :param tag: Tag value to remove
+        :param tag: Dictionary describing the tag(s) to remove (values are ignored)
         """
 
         self.backend.remove_tag(stream_id, tag)
@@ -359,7 +361,7 @@ class Datastream(object):
         Finds all streams matching the specified query tags.
 
         :param query_tags: Tags that should be matched to streams
-        :return: A list of matched stream descriptors
+        :return: A `Streams` iterator over matched stream descriptors
         """
 
         return self.backend.find_streams(query_tags)
@@ -382,7 +384,7 @@ class Datastream(object):
 
         return self.backend.append(stream_id, value, timestamp, check_timestamp)
 
-    def get_data(self, stream_id, granularity, start=None, end=None, start_exclusive=None, end_exclusive=None, reverse=False, value_downsamplers=None, time_downsamplers=None):
+    def get_data(self, stream_id, granularity, start=None, end=None, start_exclusive=None, end_exclusive=None, reverse=False, value_downsamplers=None, time_downsamplers=None, limit=None):
         """
         Retrieves data from a certain time range and of a certain granularity.
 
@@ -395,6 +397,7 @@ class Datastream(object):
         :param reverse: Should datapoints be returned in oldest to newest order (false), or in reverse (true)
         :param value_downsamplers: The list of downsamplers to limit datapoint values to (optional)
         :param time_downsamplers: The list of downsamplers to limit timestamp values to (optional)
+        :param limit: Limit the number of returned datapoints (optional)
         :return: A `Datapoints` iterator over datapoints
         """
 
@@ -431,9 +434,15 @@ class Datastream(object):
             if len(unsupported_downsamplers) > 0:
                 raise exceptions.UnsupportedDownsampler("Unsupported time downsampler(s): %s" % unsupported_downsamplers)
 
-        return self.backend.get_data(stream_id, granularity, start, end, start_exclusive, end_exclusive, reverse, value_downsamplers, time_downsamplers)
+        if limit is not None:
+            try:
+                limit = int(limit)
+            except ValueError:
+                raise ValueError("Limit value must be an integer!")
 
-    def downsample_streams(self, query_tags=None, until=None):
+        return self.backend.get_data(stream_id, granularity, start, end, start_exclusive, end_exclusive, reverse, value_downsamplers, time_downsamplers, limit)
+
+    def downsample_streams(self, query_tags=None, until=None, return_datapoints=False):
         """
         Requests the backend to downsample all streams matching the specified
         query tags. Once a time range has been downsampled, new datapoints
@@ -442,14 +451,17 @@ class Datastream(object):
         :param query_tags: Tags that should be matched to streams
         :param until: Timestamp until which to downsample, not including datapoints
                       at a timestamp (optional, otherwise all until the current time)
+        :param return_datapoints: Should newly downsampled datapoints be returned, this can
+                                  potentially create a huge temporary list and memory consumption
+                                  when downsampling many streams and datapoints
         :return: A list of dictionaries containing `stream_id`, `granularity`, and `datapoint`
-                 for each datapoint created while downsampling
+                 for each datapoint created while downsampling, if `return_datapoints` was set
         """
 
         if until is not None and until.tzinfo is None:
             until = until.replace(tzinfo=pytz.utc)
 
-        return self.backend.downsample_streams(query_tags, until)
+        return self.backend.downsample_streams(query_tags, until, return_datapoints)
 
     def backprocess_streams(self, query_tags=None):
         """
@@ -463,7 +475,7 @@ class Datastream(object):
     def delete_streams(self, query_tags=None):
         """
         Deletes datapoints for all streams matching the specified
-        query tags. If no query tags are specified, all downstream-related
+        query tags. If no query tags are specified, all datastream-related
         data is deleted from the backend.
 
         :param query_tags: Tags that should be matched to streams
